/*
 * This file was generated automatically by ExtUtils::ParseXS version 3.45 from the
 * contents of Settings.xs. Do not edit this file, edit Settings.xs instead.
 *
 *    ANY CHANGES MADE HERE WILL BE LOST!
 *
 */

#line 1 "Settings.xs"
#define PERL_NO_GET_CONTEXT
#include "module.h"
#include <irssi/src/core/misc.h>

static GHashTable *perl_settings;

static void perl_settings_add(const char *key)
{
	PERL_SCRIPT_REC *script;
        GSList *list;

	script = perl_script_find_package(perl_get_package());
	g_return_if_fail(script != NULL);

	list = g_hash_table_lookup(perl_settings, script);
        list = g_slist_append(list, g_strdup(key));
	g_hash_table_insert(perl_settings, script, list);
}

static void perl_settings_remove(const char *key)
{
	PERL_SCRIPT_REC *script;
        GSList *list, *pos;

	script = perl_script_find_package(perl_get_package());
	g_return_if_fail(script != NULL);

	list = g_hash_table_lookup(perl_settings, script);
	pos = i_slist_find_icase_string(list, key);
	if (pos != NULL) {
		list = g_slist_remove(list, pos->data);
		g_hash_table_insert(perl_settings, script, list);
	}
}

static void perl_settings_free(PERL_SCRIPT_REC *script, GSList *list)
{
	g_slist_foreach(list, (GFunc) g_free, NULL);
        g_slist_free(list);
}

static void sig_script_destroyed(PERL_SCRIPT_REC *script)
{
	GSList *list;

	list = g_hash_table_lookup(perl_settings, script);
	if (list != NULL) {
                g_slist_foreach(list, (GFunc) settings_remove, NULL);
		perl_settings_free(script, list);
		g_hash_table_remove(perl_settings, script);
	}
}

void perl_settings_init(void)
{
	perl_settings = g_hash_table_new((GHashFunc) g_direct_hash,
					 (GCompareFunc) g_direct_equal);
        signal_add("script destroyed", (SIGNAL_FUNC) sig_script_destroyed);
}

void perl_settings_deinit(void)
{
        signal_remove("script destroyed", (SIGNAL_FUNC) sig_script_destroyed);

	g_hash_table_foreach(perl_settings, (GHFunc) perl_settings_free, NULL);
	g_hash_table_destroy(perl_settings);
}

#line 79 "Settings.c"
#ifndef PERL_UNUSED_VAR
#  define PERL_UNUSED_VAR(var) if (0) var = var
#endif

#ifndef dVAR
#  define dVAR		dNOOP
#endif


/* This stuff is not part of the API! You have been warned. */
#ifndef PERL_VERSION_DECIMAL
#  define PERL_VERSION_DECIMAL(r,v,s) (r*1000000 + v*1000 + s)
#endif
#ifndef PERL_DECIMAL_VERSION
#  define PERL_DECIMAL_VERSION \
	  PERL_VERSION_DECIMAL(PERL_REVISION,PERL_VERSION,PERL_SUBVERSION)
#endif
#ifndef PERL_VERSION_GE
#  define PERL_VERSION_GE(r,v,s) \
	  (PERL_DECIMAL_VERSION >= PERL_VERSION_DECIMAL(r,v,s))
#endif
#ifndef PERL_VERSION_LE
#  define PERL_VERSION_LE(r,v,s) \
	  (PERL_DECIMAL_VERSION <= PERL_VERSION_DECIMAL(r,v,s))
#endif

/* XS_INTERNAL is the explicit static-linkage variant of the default
 * XS macro.
 *
 * XS_EXTERNAL is the same as XS_INTERNAL except it does not include
 * "STATIC", ie. it exports XSUB symbols. You probably don't want that
 * for anything but the BOOT XSUB.
 *
 * See XSUB.h in core!
 */


/* TODO: This might be compatible further back than 5.10.0. */
#if PERL_VERSION_GE(5, 10, 0) && PERL_VERSION_LE(5, 15, 1)
#  undef XS_EXTERNAL
#  undef XS_INTERNAL
#  if defined(__CYGWIN__) && defined(USE_DYNAMIC_LOADING)
#    define XS_EXTERNAL(name) __declspec(dllexport) XSPROTO(name)
#    define XS_INTERNAL(name) STATIC XSPROTO(name)
#  endif
#  if defined(__SYMBIAN32__)
#    define XS_EXTERNAL(name) EXPORT_C XSPROTO(name)
#    define XS_INTERNAL(name) EXPORT_C STATIC XSPROTO(name)
#  endif
#  ifndef XS_EXTERNAL
#    if defined(HASATTRIBUTE_UNUSED) && !defined(__cplusplus)
#      define XS_EXTERNAL(name) void name(pTHX_ CV* cv __attribute__unused__)
#      define XS_INTERNAL(name) STATIC void name(pTHX_ CV* cv __attribute__unused__)
#    else
#      ifdef __cplusplus
#        define XS_EXTERNAL(name) extern "C" XSPROTO(name)
#        define XS_INTERNAL(name) static XSPROTO(name)
#      else
#        define XS_EXTERNAL(name) XSPROTO(name)
#        define XS_INTERNAL(name) STATIC XSPROTO(name)
#      endif
#    endif
#  endif
#endif

/* perl >= 5.10.0 && perl <= 5.15.1 */


/* The XS_EXTERNAL macro is used for functions that must not be static
 * like the boot XSUB of a module. If perl didn't have an XS_EXTERNAL
 * macro defined, the best we can do is assume XS is the same.
 * Dito for XS_INTERNAL.
 */
#ifndef XS_EXTERNAL
#  define XS_EXTERNAL(name) XS(name)
#endif
#ifndef XS_INTERNAL
#  define XS_INTERNAL(name) XS(name)
#endif

/* Now, finally, after all this mess, we want an ExtUtils::ParseXS
 * internal macro that we're free to redefine for varying linkage due
 * to the EXPORT_XSUB_SYMBOLS XS keyword. This is internal, use
 * XS_EXTERNAL(name) or XS_INTERNAL(name) in your code if you need to!
 */

#undef XS_EUPXS
#if defined(PERL_EUPXS_ALWAYS_EXPORT)
#  define XS_EUPXS(name) XS_EXTERNAL(name)
#else
   /* default to internal */
#  define XS_EUPXS(name) XS_INTERNAL(name)
#endif

#ifndef PERL_ARGS_ASSERT_CROAK_XS_USAGE
#define PERL_ARGS_ASSERT_CROAK_XS_USAGE assert(cv); assert(params)

/* prototype to pass -Wmissing-prototypes */
STATIC void
S_croak_xs_usage(const CV *const cv, const char *const params);

STATIC void
S_croak_xs_usage(const CV *const cv, const char *const params)
{
    const GV *const gv = CvGV(cv);

    PERL_ARGS_ASSERT_CROAK_XS_USAGE;

    if (gv) {
        const char *const gvname = GvNAME(gv);
        const HV *const stash = GvSTASH(gv);
        const char *const hvname = stash ? HvNAME(stash) : NULL;

        if (hvname)
	    Perl_croak_nocontext("Usage: %s::%s(%s)", hvname, gvname, params);
        else
	    Perl_croak_nocontext("Usage: %s(%s)", gvname, params);
    } else {
        /* Pants. I don't think that it should be possible to get here. */
	Perl_croak_nocontext("Usage: CODE(0x%" UVxf ")(%s)", PTR2UV(cv), params);
    }
}
#undef  PERL_ARGS_ASSERT_CROAK_XS_USAGE

#define croak_xs_usage        S_croak_xs_usage

#endif

/* NOTE: the prototype of newXSproto() is different in versions of perls,
 * so we define a portable version of newXSproto()
 */
#ifdef newXS_flags
#define newXSproto_portable(name, c_impl, file, proto) newXS_flags(name, c_impl, file, proto, 0)
#else
#define newXSproto_portable(name, c_impl, file, proto) (PL_Sv=(SV*)newXS(name, c_impl, file), sv_setpv(PL_Sv, proto), (CV*)PL_Sv)
#endif /* !defined(newXS_flags) */

#if PERL_VERSION_LE(5, 21, 5)
#  define newXS_deffile(a,b) Perl_newXS(aTHX_ a,b,file)
#else
#  define newXS_deffile(a,b) Perl_newXS_deffile(aTHX_ a,b)
#endif

#line 223 "Settings.c"

XS_EUPXS(XS_Irssi_settings_get_str); /* prototype to pass -Wmissing-prototypes */
XS_EUPXS(XS_Irssi_settings_get_str)
{
    dVAR; dXSARGS;
    if (items != 1)
       croak_xs_usage(cv,  "key");
    {
	char *	key = (char *)SvPV_nolen(ST(0))
;
#line 76 "Settings.xs"
	const char *str;
#line 236 "Settings.c"
	SV *	RETVAL;
#line 78 "Settings.xs"
	str = settings_get_str(key);
	RETVAL = new_pv(str);
#line 241 "Settings.c"
	RETVAL = sv_2mortal(RETVAL);
	ST(0) = RETVAL;
    }
    XSRETURN(1);
}


XS_EUPXS(XS_Irssi_settings_get_int); /* prototype to pass -Wmissing-prototypes */
XS_EUPXS(XS_Irssi_settings_get_int)
{
    dVAR; dXSARGS;
    if (items != 1)
       croak_xs_usage(cv,  "key");
    {
	char *	key = (char *)SvPV_nolen(ST(0))
;
	int	RETVAL;
	dXSTARG;

	RETVAL = settings_get_int(key);
	XSprePUSH;
	PUSHi((IV)RETVAL);
    }
    XSRETURN(1);
}


XS_EUPXS(XS_Irssi_settings_get_bool); /* prototype to pass -Wmissing-prototypes */
XS_EUPXS(XS_Irssi_settings_get_bool)
{
    dVAR; dXSARGS;
    if (items != 1)
       croak_xs_usage(cv,  "key");
    {
	char *	key = (char *)SvPV_nolen(ST(0))
;
	int	RETVAL;
	dXSTARG;

	RETVAL = settings_get_bool(key);
	XSprePUSH;
	PUSHi((IV)RETVAL);
    }
    XSRETURN(1);
}


XS_EUPXS(XS_Irssi_settings_get_time); /* prototype to pass -Wmissing-prototypes */
XS_EUPXS(XS_Irssi_settings_get_time)
{
    dVAR; dXSARGS;
    if (items != 1)
       croak_xs_usage(cv,  "key");
    {
	char *	key = (char *)SvPV_nolen(ST(0))
;
	int	RETVAL;
	dXSTARG;

	RETVAL = settings_get_time(key);
	XSprePUSH;
	PUSHi((IV)RETVAL);
    }
    XSRETURN(1);
}


XS_EUPXS(XS_Irssi_settings_get_level); /* prototype to pass -Wmissing-prototypes */
XS_EUPXS(XS_Irssi_settings_get_level)
{
    dVAR; dXSARGS;
    if (items != 1)
       croak_xs_usage(cv,  "key");
    {
	char *	key = (char *)SvPV_nolen(ST(0))
;
	int	RETVAL;
	dXSTARG;

	RETVAL = settings_get_level(key);
	XSprePUSH;
	PUSHi((IV)RETVAL);
    }
    XSRETURN(1);
}


XS_EUPXS(XS_Irssi_settings_get_level_negative); /* prototype to pass -Wmissing-prototypes */
XS_EUPXS(XS_Irssi_settings_get_level_negative)
{
    dVAR; dXSARGS;
    if (items != 1)
       croak_xs_usage(cv,  "key");
    {
	char *	key = (char *)SvPV_nolen(ST(0))
;
	int	RETVAL;
	dXSTARG;

	RETVAL = settings_get_level_negative(key);
	XSprePUSH;
	PUSHi((IV)RETVAL);
    }
    XSRETURN(1);
}


XS_EUPXS(XS_Irssi_settings_get_size); /* prototype to pass -Wmissing-prototypes */
XS_EUPXS(XS_Irssi_settings_get_size)
{
    dVAR; dXSARGS;
    if (items != 1)
       croak_xs_usage(cv,  "key");
    {
	char *	key = (char *)SvPV_nolen(ST(0))
;
	int	RETVAL;
	dXSTARG;

	RETVAL = settings_get_size(key);
	XSprePUSH;
	PUSHi((IV)RETVAL);
    }
    XSRETURN(1);
}


XS_EUPXS(XS_Irssi_settings_get_choice); /* prototype to pass -Wmissing-prototypes */
XS_EUPXS(XS_Irssi_settings_get_choice)
{
    dVAR; dXSARGS;
    if (items != 1)
       croak_xs_usage(cv,  "key");
    {
	char *	key = (char *)SvPV_nolen(ST(0))
;
	int	RETVAL;
	dXSTARG;

	RETVAL = settings_get_choice(key);
	XSprePUSH;
	PUSHi((IV)RETVAL);
    }
    XSRETURN(1);
}


XS_EUPXS(XS_Irssi_settings_set_str); /* prototype to pass -Wmissing-prototypes */
XS_EUPXS(XS_Irssi_settings_set_str)
{
    dVAR; dXSARGS;
    if (items != 2)
       croak_xs_usage(cv,  "key, value");
    {
	char *	key = (char *)SvPV_nolen(ST(0))
;
	char *	value = (char *)SvPV_nolen(ST(1))
;

	settings_set_str(key, value);
    }
    XSRETURN_EMPTY;
}


XS_EUPXS(XS_Irssi_settings_set_int); /* prototype to pass -Wmissing-prototypes */
XS_EUPXS(XS_Irssi_settings_set_int)
{
    dVAR; dXSARGS;
    if (items != 2)
       croak_xs_usage(cv,  "key, value");
    {
	char *	key = (char *)SvPV_nolen(ST(0))
;
	int	value = (int)SvIV(ST(1))
;

	settings_set_int(key, value);
    }
    XSRETURN_EMPTY;
}


XS_EUPXS(XS_Irssi_settings_set_bool); /* prototype to pass -Wmissing-prototypes */
XS_EUPXS(XS_Irssi_settings_set_bool)
{
    dVAR; dXSARGS;
    if (items != 2)
       croak_xs_usage(cv,  "key, value");
    {
	char *	key = (char *)SvPV_nolen(ST(0))
;
	int	value = (int)SvIV(ST(1))
;

	settings_set_bool(key, value);
    }
    XSRETURN_EMPTY;
}


XS_EUPXS(XS_Irssi_settings_set_time); /* prototype to pass -Wmissing-prototypes */
XS_EUPXS(XS_Irssi_settings_set_time)
{
    dVAR; dXSARGS;
    if (items != 2)
       croak_xs_usage(cv,  "key, value");
    {
	char *	key = (char *)SvPV_nolen(ST(0))
;
	char *	value = (char *)SvPV_nolen(ST(1))
;
	int	RETVAL;
	dXSTARG;

	RETVAL = settings_set_time(key, value);
	XSprePUSH;
	PUSHi((IV)RETVAL);
    }
    XSRETURN(1);
}


XS_EUPXS(XS_Irssi_settings_set_level); /* prototype to pass -Wmissing-prototypes */
XS_EUPXS(XS_Irssi_settings_set_level)
{
    dVAR; dXSARGS;
    if (items != 2)
       croak_xs_usage(cv,  "key, value");
    {
	char *	key = (char *)SvPV_nolen(ST(0))
;
	char *	value = (char *)SvPV_nolen(ST(1))
;
	int	RETVAL;
	dXSTARG;

	RETVAL = settings_set_level(key, value);
	XSprePUSH;
	PUSHi((IV)RETVAL);
    }
    XSRETURN(1);
}


XS_EUPXS(XS_Irssi_settings_set_size); /* prototype to pass -Wmissing-prototypes */
XS_EUPXS(XS_Irssi_settings_set_size)
{
    dVAR; dXSARGS;
    if (items != 2)
       croak_xs_usage(cv,  "key, value");
    {
	char *	key = (char *)SvPV_nolen(ST(0))
;
	char *	value = (char *)SvPV_nolen(ST(1))
;
	int	RETVAL;
	dXSTARG;

	RETVAL = settings_set_size(key, value);
	XSprePUSH;
	PUSHi((IV)RETVAL);
    }
    XSRETURN(1);
}


XS_EUPXS(XS_Irssi_settings_set_choice); /* prototype to pass -Wmissing-prototypes */
XS_EUPXS(XS_Irssi_settings_set_choice)
{
    dVAR; dXSARGS;
    if (items != 2)
       croak_xs_usage(cv,  "key, value");
    {
	char *	key = (char *)SvPV_nolen(ST(0))
;
	char *	value = (char *)SvPV_nolen(ST(1))
;
	int	RETVAL;
	dXSTARG;

	RETVAL = settings_set_choice(key, value);
	XSprePUSH;
	PUSHi((IV)RETVAL);
    }
    XSRETURN(1);
}


XS_EUPXS(XS_Irssi_settings_add_str); /* prototype to pass -Wmissing-prototypes */
XS_EUPXS(XS_Irssi_settings_add_str)
{
    dVAR; dXSARGS;
    if (items != 3)
       croak_xs_usage(cv,  "section, key, def");
    {
	char *	section = (char *)SvPV_nolen(ST(0))
;
	char *	key = (char *)SvPV_nolen(ST(1))
;
	char *	def = (char *)SvPV_nolen(ST(2))
;
#line 152 "Settings.xs"
        perl_settings_add(key);
	settings_add_str_module(MODULE_NAME"/scripts", section, key, def);
#line 547 "Settings.c"
    }
    XSRETURN_EMPTY;
}


XS_EUPXS(XS_Irssi_settings_add_int); /* prototype to pass -Wmissing-prototypes */
XS_EUPXS(XS_Irssi_settings_add_int)
{
    dVAR; dXSARGS;
    if (items != 3)
       croak_xs_usage(cv,  "section, key, def");
    {
	char *	section = (char *)SvPV_nolen(ST(0))
;
	char *	key = (char *)SvPV_nolen(ST(1))
;
	int	def = (int)SvIV(ST(2))
;
#line 161 "Settings.xs"
        perl_settings_add(key);
	settings_add_int_module(MODULE_NAME"/scripts", section, key, def);
#line 569 "Settings.c"
    }
    XSRETURN_EMPTY;
}


XS_EUPXS(XS_Irssi_settings_add_bool); /* prototype to pass -Wmissing-prototypes */
XS_EUPXS(XS_Irssi_settings_add_bool)
{
    dVAR; dXSARGS;
    if (items != 3)
       croak_xs_usage(cv,  "section, key, def");
    {
	char *	section = (char *)SvPV_nolen(ST(0))
;
	char *	key = (char *)SvPV_nolen(ST(1))
;
	int	def = (int)SvIV(ST(2))
;
#line 170 "Settings.xs"
        perl_settings_add(key);
	settings_add_bool_module(MODULE_NAME"/scripts", section, key, def);
#line 591 "Settings.c"
    }
    XSRETURN_EMPTY;
}


XS_EUPXS(XS_Irssi_settings_add_time); /* prototype to pass -Wmissing-prototypes */
XS_EUPXS(XS_Irssi_settings_add_time)
{
    dVAR; dXSARGS;
    if (items != 3)
       croak_xs_usage(cv,  "section, key, def");
    {
	char *	section = (char *)SvPV_nolen(ST(0))
;
	char *	key = (char *)SvPV_nolen(ST(1))
;
	char *	def = (char *)SvPV_nolen(ST(2))
;
#line 179 "Settings.xs"
        perl_settings_add(key);
	settings_add_time_module(MODULE_NAME"/scripts", section, key, def);
#line 613 "Settings.c"
    }
    XSRETURN_EMPTY;
}


XS_EUPXS(XS_Irssi_settings_add_level); /* prototype to pass -Wmissing-prototypes */
XS_EUPXS(XS_Irssi_settings_add_level)
{
    dVAR; dXSARGS;
    if (items != 3)
       croak_xs_usage(cv,  "section, key, def");
    {
	char *	section = (char *)SvPV_nolen(ST(0))
;
	char *	key = (char *)SvPV_nolen(ST(1))
;
	char *	def = (char *)SvPV_nolen(ST(2))
;
#line 188 "Settings.xs"
        perl_settings_add(key);
	settings_add_level_module(MODULE_NAME"/scripts", section, key, def);
#line 635 "Settings.c"
    }
    XSRETURN_EMPTY;
}


XS_EUPXS(XS_Irssi_settings_add_size); /* prototype to pass -Wmissing-prototypes */
XS_EUPXS(XS_Irssi_settings_add_size)
{
    dVAR; dXSARGS;
    if (items != 3)
       croak_xs_usage(cv,  "section, key, def");
    {
	char *	section = (char *)SvPV_nolen(ST(0))
;
	char *	key = (char *)SvPV_nolen(ST(1))
;
	char *	def = (char *)SvPV_nolen(ST(2))
;
#line 197 "Settings.xs"
        perl_settings_add(key);
	settings_add_size_module(MODULE_NAME"/scripts", section, key, def);
#line 657 "Settings.c"
    }
    XSRETURN_EMPTY;
}


XS_EUPXS(XS_Irssi_settings_add_choice); /* prototype to pass -Wmissing-prototypes */
XS_EUPXS(XS_Irssi_settings_add_choice)
{
    dVAR; dXSARGS;
    if (items != 4)
       croak_xs_usage(cv,  "section, key, def, choices");
    {
	char *	section = (char *)SvPV_nolen(ST(0))
;
	char *	key = (char *)SvPV_nolen(ST(1))
;
	int	def = (int)SvIV(ST(2))
;
	char *	choices = (char *)SvPV_nolen(ST(3))
;
#line 207 "Settings.xs"
	perl_settings_add(key);
	settings_add_choice_module(MODULE_NAME "/scripts", section, key, def, choices);
#line 681 "Settings.c"
    }
    XSRETURN_EMPTY;
}


XS_EUPXS(XS_Irssi_settings_remove); /* prototype to pass -Wmissing-prototypes */
XS_EUPXS(XS_Irssi_settings_remove)
{
    dVAR; dXSARGS;
    if (items != 1)
       croak_xs_usage(cv,  "key");
    {
	char *	key = (char *)SvPV_nolen(ST(0))
;
#line 214 "Settings.xs"
        perl_settings_remove(key);
	settings_remove(key);
#line 699 "Settings.c"
    }
    XSRETURN_EMPTY;
}

#ifdef __cplusplus
extern "C"
#endif
XS_EXTERNAL(boot_Irssi__Settings); /* prototype to pass -Wmissing-prototypes */
XS_EXTERNAL(boot_Irssi__Settings)
{
#if PERL_VERSION_LE(5, 21, 5)
    dVAR; dXSARGS;
#else
    dVAR; dXSBOOTARGSXSAPIVERCHK;
#endif
#if PERL_VERSION_LE(5, 8, 999) /* PERL_VERSION_LT is 5.33+ */
    char* file = __FILE__;
#else
    const char* file = __FILE__;
#endif

    PERL_UNUSED_VAR(file);

    PERL_UNUSED_VAR(cv); /* -W */
    PERL_UNUSED_VAR(items); /* -W */
#if PERL_VERSION_LE(5, 21, 5)
    XS_VERSION_BOOTCHECK;
#  ifdef XS_APIVERSION_BOOTCHECK
    XS_APIVERSION_BOOTCHECK;
#  endif
#endif

        (void)newXSproto_portable("Irssi::settings_get_str", XS_Irssi_settings_get_str, file, "$");
        (void)newXSproto_portable("Irssi::settings_get_int", XS_Irssi_settings_get_int, file, "$");
        (void)newXSproto_portable("Irssi::settings_get_bool", XS_Irssi_settings_get_bool, file, "$");
        (void)newXSproto_portable("Irssi::settings_get_time", XS_Irssi_settings_get_time, file, "$");
        (void)newXSproto_portable("Irssi::settings_get_level", XS_Irssi_settings_get_level, file, "$");
        (void)newXSproto_portable("Irssi::settings_get_level_negative", XS_Irssi_settings_get_level_negative, file, "$");
        (void)newXSproto_portable("Irssi::settings_get_size", XS_Irssi_settings_get_size, file, "$");
        (void)newXSproto_portable("Irssi::settings_get_choice", XS_Irssi_settings_get_choice, file, "$");
        (void)newXSproto_portable("Irssi::settings_set_str", XS_Irssi_settings_set_str, file, "$$");
        (void)newXSproto_portable("Irssi::settings_set_int", XS_Irssi_settings_set_int, file, "$$");
        (void)newXSproto_portable("Irssi::settings_set_bool", XS_Irssi_settings_set_bool, file, "$$");
        (void)newXSproto_portable("Irssi::settings_set_time", XS_Irssi_settings_set_time, file, "$$");
        (void)newXSproto_portable("Irssi::settings_set_level", XS_Irssi_settings_set_level, file, "$$");
        (void)newXSproto_portable("Irssi::settings_set_size", XS_Irssi_settings_set_size, file, "$$");
        (void)newXSproto_portable("Irssi::settings_set_choice", XS_Irssi_settings_set_choice, file, "$$");
        (void)newXSproto_portable("Irssi::settings_add_str", XS_Irssi_settings_add_str, file, "$$$");
        (void)newXSproto_portable("Irssi::settings_add_int", XS_Irssi_settings_add_int, file, "$$$");
        (void)newXSproto_portable("Irssi::settings_add_bool", XS_Irssi_settings_add_bool, file, "$$$");
        (void)newXSproto_portable("Irssi::settings_add_time", XS_Irssi_settings_add_time, file, "$$$");
        (void)newXSproto_portable("Irssi::settings_add_level", XS_Irssi_settings_add_level, file, "$$$");
        (void)newXSproto_portable("Irssi::settings_add_size", XS_Irssi_settings_add_size, file, "$$$");
        (void)newXSproto_portable("Irssi::settings_add_choice", XS_Irssi_settings_add_choice, file, "$$$$");
        (void)newXSproto_portable("Irssi::settings_remove", XS_Irssi_settings_remove, file, "$");
#if PERL_VERSION_LE(5, 21, 5)
#  if PERL_VERSION_GE(5, 9, 0)
    if (PL_unitcheckav)
        call_list(PL_scopestack_ix, PL_unitcheckav);
#  endif
    XSRETURN_YES;
#else
    Perl_xs_boot_epilog(aTHX_ ax);
#endif
}

