/*
 * This file was generated automatically by ExtUtils::ParseXS version 3.45 from the
 * contents of UI.xs. Do not edit this file, edit UI.xs instead.
 *
 *    ANY CHANGES MADE HERE WILL BE LOST!
 *
 */

#line 1 "UI.xs"
#define PERL_NO_GET_CONTEXT
#include "module.h"

void perl_themes_init(void);
void perl_themes_deinit(void);

static int initialized = FALSE;

static void perl_process_fill_hash(HV *hv, PROCESS_REC *process)
{
	(void) hv_store(hv, "id", 2, newSViv(process->id), 0);
	(void) hv_store(hv, "name", 4, new_pv(process->name), 0);
	(void) hv_store(hv, "args", 4, new_pv(process->args), 0);

	(void) hv_store(hv, "pid", 3, newSViv(process->pid), 0);
	(void) hv_store(hv, "target", 6, new_pv(process->target), 0);
	if (process->target_win != NULL) {
		(void) hv_store(hv, "target_win", 10,
			 plain_bless(process->target_win, "Irssi::UI::Window"), 0);
	}
	(void) hv_store(hv, "shell", 5, newSViv(process->shell), 0);
	(void) hv_store(hv, "notice", 6, newSViv(process->notice), 0);
	(void) hv_store(hv, "silent", 6, newSViv(process->silent), 0);
}

static void perl_window_fill_hash(HV *hv, WINDOW_REC *window)
{
	(void) hv_store(hv, "refnum", 6, newSViv(window->refnum), 0);
	(void) hv_store(hv, "name", 4, new_pv(window->name), 0);
	(void) hv_store(hv, "history_name", 12, new_pv(window->history_name), 0);

	(void) hv_store(hv, "width", 5, newSViv(window->width), 0);
	(void) hv_store(hv, "height", 6, newSViv(window->height), 0);

	if (window->active)
		(void) hv_store(hv, "active", 6, iobject_bless(window->active), 0);
	if (window->active_server)
		(void) hv_store(hv, "active_server", 13, iobject_bless(window->active_server), 0);

	(void) hv_store(hv, "servertag", 9, new_pv(window->servertag), 0);
	(void) hv_store(hv, "level", 5, newSViv(window->level), 0);

	(void) hv_store(hv, "immortal", 8, newSViv(window->immortal), 0);
	(void) hv_store(hv, "sticky_refnum", 13, newSViv(window->sticky_refnum), 0);

	(void) hv_store(hv, "data_level", 10, newSViv(window->data_level), 0);
	(void) hv_store(hv, "hilight_color", 13, new_pv(window->hilight_color), 0);

	(void) hv_store(hv, "last_timestamp", 14, newSViv(window->last_timestamp), 0);
	(void) hv_store(hv, "last_line", 9, newSViv(window->last_line), 0);

	(void) hv_store(hv, "theme", 5, plain_bless(window->theme, "Irssi::UI::Theme"), 0);
	(void) hv_store(hv, "theme_name", 10, new_pv(window->theme_name), 0);
}

static void perl_text_dest_fill_hash(HV *hv, TEXT_DEST_REC *dest)
{
	(void) hv_store(hv, "window", 6, plain_bless(dest->window, "Irssi::UI::Window"), 0);
	(void) hv_store(hv, "server", 6, iobject_bless(dest->server), 0);
	(void) hv_store(hv, "target", 6, new_pv(dest->target), 0);
	(void) hv_store(hv, "level", 5, newSViv(dest->level), 0);

	(void) hv_store(hv, "hilight_priority", 16, newSViv(dest->hilight_priority), 0);
	(void) hv_store(hv, "hilight_color", 13, new_pv(dest->hilight_color), 0);
}

static void perl_line_info_meta_fill_hash(HV *hv, LINE_INFO_META_REC *meta)
{
	GHashTableIter iter;
	char *key;
	char *val;

	if (meta != NULL) {
		if (meta->hash != NULL) {
			g_hash_table_iter_init(&iter, meta->hash);
			while (
			    g_hash_table_iter_next(&iter, (gpointer *) &key, (gpointer *) &val)) {
				(void) hv_store(hv, key, strlen(key), new_pv(val), 0);
			}
		}
		if (meta->server_time) {
			(void) hv_store(hv, "server_time", 11, newSViv(meta->server_time), 0);
		}
	}
}

static void perl_exec_fill_hash(HV *hv, EXEC_WI_REC *item)
{
	g_return_if_fail(hv != NULL);
	g_return_if_fail(item != NULL);

	perl_window_item_fill_hash(hv, (WI_ITEM_REC *) item);
	/* we don't bless to Process here to avoid infinite recursion
	   in the simplistic script binding */
	if (item->process != NULL) {
		(void) hv_store(hv, "process_id", 10, newSViv(item->process->id), 0);
	}
}

static PLAIN_OBJECT_INIT_REC fe_plains[] = {
	{ "Irssi::UI::Process", (PERL_OBJECT_FUNC) perl_process_fill_hash },
	{ "Irssi::UI::Window", (PERL_OBJECT_FUNC) perl_window_fill_hash },
	{ "Irssi::UI::TextDest", (PERL_OBJECT_FUNC) perl_text_dest_fill_hash },
	{ "Irssi::UI::LineInfoMeta", (PERL_OBJECT_FUNC) perl_line_info_meta_fill_hash },

	{ NULL, NULL }
};

#line 119 "UI.c"
#ifndef PERL_UNUSED_VAR
#  define PERL_UNUSED_VAR(var) if (0) var = var
#endif

#ifndef dVAR
#  define dVAR		dNOOP
#endif


/* This stuff is not part of the API! You have been warned. */
#ifndef PERL_VERSION_DECIMAL
#  define PERL_VERSION_DECIMAL(r,v,s) (r*1000000 + v*1000 + s)
#endif
#ifndef PERL_DECIMAL_VERSION
#  define PERL_DECIMAL_VERSION \
	  PERL_VERSION_DECIMAL(PERL_REVISION,PERL_VERSION,PERL_SUBVERSION)
#endif
#ifndef PERL_VERSION_GE
#  define PERL_VERSION_GE(r,v,s) \
	  (PERL_DECIMAL_VERSION >= PERL_VERSION_DECIMAL(r,v,s))
#endif
#ifndef PERL_VERSION_LE
#  define PERL_VERSION_LE(r,v,s) \
	  (PERL_DECIMAL_VERSION <= PERL_VERSION_DECIMAL(r,v,s))
#endif

/* XS_INTERNAL is the explicit static-linkage variant of the default
 * XS macro.
 *
 * XS_EXTERNAL is the same as XS_INTERNAL except it does not include
 * "STATIC", ie. it exports XSUB symbols. You probably don't want that
 * for anything but the BOOT XSUB.
 *
 * See XSUB.h in core!
 */


/* TODO: This might be compatible further back than 5.10.0. */
#if PERL_VERSION_GE(5, 10, 0) && PERL_VERSION_LE(5, 15, 1)
#  undef XS_EXTERNAL
#  undef XS_INTERNAL
#  if defined(__CYGWIN__) && defined(USE_DYNAMIC_LOADING)
#    define XS_EXTERNAL(name) __declspec(dllexport) XSPROTO(name)
#    define XS_INTERNAL(name) STATIC XSPROTO(name)
#  endif
#  if defined(__SYMBIAN32__)
#    define XS_EXTERNAL(name) EXPORT_C XSPROTO(name)
#    define XS_INTERNAL(name) EXPORT_C STATIC XSPROTO(name)
#  endif
#  ifndef XS_EXTERNAL
#    if defined(HASATTRIBUTE_UNUSED) && !defined(__cplusplus)
#      define XS_EXTERNAL(name) void name(pTHX_ CV* cv __attribute__unused__)
#      define XS_INTERNAL(name) STATIC void name(pTHX_ CV* cv __attribute__unused__)
#    else
#      ifdef __cplusplus
#        define XS_EXTERNAL(name) extern "C" XSPROTO(name)
#        define XS_INTERNAL(name) static XSPROTO(name)
#      else
#        define XS_EXTERNAL(name) XSPROTO(name)
#        define XS_INTERNAL(name) STATIC XSPROTO(name)
#      endif
#    endif
#  endif
#endif

/* perl >= 5.10.0 && perl <= 5.15.1 */


/* The XS_EXTERNAL macro is used for functions that must not be static
 * like the boot XSUB of a module. If perl didn't have an XS_EXTERNAL
 * macro defined, the best we can do is assume XS is the same.
 * Dito for XS_INTERNAL.
 */
#ifndef XS_EXTERNAL
#  define XS_EXTERNAL(name) XS(name)
#endif
#ifndef XS_INTERNAL
#  define XS_INTERNAL(name) XS(name)
#endif

/* Now, finally, after all this mess, we want an ExtUtils::ParseXS
 * internal macro that we're free to redefine for varying linkage due
 * to the EXPORT_XSUB_SYMBOLS XS keyword. This is internal, use
 * XS_EXTERNAL(name) or XS_INTERNAL(name) in your code if you need to!
 */

#undef XS_EUPXS
#if defined(PERL_EUPXS_ALWAYS_EXPORT)
#  define XS_EUPXS(name) XS_EXTERNAL(name)
#else
   /* default to internal */
#  define XS_EUPXS(name) XS_INTERNAL(name)
#endif

#ifndef PERL_ARGS_ASSERT_CROAK_XS_USAGE
#define PERL_ARGS_ASSERT_CROAK_XS_USAGE assert(cv); assert(params)

/* prototype to pass -Wmissing-prototypes */
STATIC void
S_croak_xs_usage(const CV *const cv, const char *const params);

STATIC void
S_croak_xs_usage(const CV *const cv, const char *const params)
{
    const GV *const gv = CvGV(cv);

    PERL_ARGS_ASSERT_CROAK_XS_USAGE;

    if (gv) {
        const char *const gvname = GvNAME(gv);
        const HV *const stash = GvSTASH(gv);
        const char *const hvname = stash ? HvNAME(stash) : NULL;

        if (hvname)
	    Perl_croak_nocontext("Usage: %s::%s(%s)", hvname, gvname, params);
        else
	    Perl_croak_nocontext("Usage: %s(%s)", gvname, params);
    } else {
        /* Pants. I don't think that it should be possible to get here. */
	Perl_croak_nocontext("Usage: CODE(0x%" UVxf ")(%s)", PTR2UV(cv), params);
    }
}
#undef  PERL_ARGS_ASSERT_CROAK_XS_USAGE

#define croak_xs_usage        S_croak_xs_usage

#endif

/* NOTE: the prototype of newXSproto() is different in versions of perls,
 * so we define a portable version of newXSproto()
 */
#ifdef newXS_flags
#define newXSproto_portable(name, c_impl, file, proto) newXS_flags(name, c_impl, file, proto, 0)
#else
#define newXSproto_portable(name, c_impl, file, proto) (PL_Sv=(SV*)newXS(name, c_impl, file), sv_setpv(PL_Sv, proto), (CV*)PL_Sv)
#endif /* !defined(newXS_flags) */

#if PERL_VERSION_LE(5, 21, 5)
#  define newXS_deffile(a,b) Perl_newXS(aTHX_ a,b,file)
#else
#  define newXS_deffile(a,b) Perl_newXS_deffile(aTHX_ a,b)
#endif

#line 263 "UI.c"

XS_EUPXS(XS_Irssi__UI_processes); /* prototype to pass -Wmissing-prototypes */
XS_EUPXS(XS_Irssi__UI_processes)
{
    dVAR; dXSARGS;
    if (items != 0)
       croak_xs_usage(cv,  "");
    PERL_UNUSED_VAR(ax); /* -Wall */
    SP -= items;
    {
#line 116 "UI.xs"
	GSList *tmp;
#line 276 "UI.c"
#line 118 "UI.xs"
	for (tmp = processes; tmp != NULL; tmp = tmp->next) {
		XPUSHs(sv_2mortal(plain_bless(tmp->data, "Irssi::UI::Process")));
	}
#line 281 "UI.c"
	PUTBACK;
	return;
    }
}


XS_EUPXS(XS_Irssi__UI_init); /* prototype to pass -Wmissing-prototypes */
XS_EUPXS(XS_Irssi__UI_init)
{
    dVAR; dXSARGS;
    if (items != 0)
       croak_xs_usage(cv,  "");
    {
#line 126 "UI.xs"
	if (initialized) return;
	perl_api_version_check("Irssi::UI");
	initialized = TRUE;

        irssi_add_plains(fe_plains);
        /* window items: fe-exec */
        irssi_add_object(module_get_uniq_id_str("WINDOW ITEM TYPE", "EXEC"),
			 0, "Irssi::UI::Exec",
                         (PERL_OBJECT_FUNC) perl_exec_fill_hash);
        perl_themes_init();
#line 306 "UI.c"
    }
    XSRETURN_EMPTY;
}


XS_EUPXS(XS_Irssi__UI_deinit); /* prototype to pass -Wmissing-prototypes */
XS_EUPXS(XS_Irssi__UI_deinit)
{
    dVAR; dXSARGS;
    if (items != 0)
       croak_xs_usage(cv,  "");
    {
#line 140 "UI.xs"
	if (!initialized) return;
        perl_themes_deinit();
	initialized = FALSE;
#line 323 "UI.c"
    }
    XSRETURN_EMPTY;
}

#ifdef __cplusplus
extern "C"
#endif
XS_EXTERNAL(boot_Irssi__UI); /* prototype to pass -Wmissing-prototypes */
XS_EXTERNAL(boot_Irssi__UI)
{
#if PERL_VERSION_LE(5, 21, 5)
    dVAR; dXSARGS;
#else
    dVAR; dXSBOOTARGSXSAPIVERCHK;
#endif
#if PERL_VERSION_LE(5, 8, 999) /* PERL_VERSION_LT is 5.33+ */
    char* file = __FILE__;
#else
    const char* file = __FILE__;
#endif

    PERL_UNUSED_VAR(file);

    PERL_UNUSED_VAR(cv); /* -W */
    PERL_UNUSED_VAR(items); /* -W */
#if PERL_VERSION_LE(5, 21, 5)
    XS_VERSION_BOOTCHECK;
#  ifdef XS_APIVERSION_BOOTCHECK
    XS_APIVERSION_BOOTCHECK;
#  endif
#endif

        (void)newXSproto_portable("Irssi::UI::processes", XS_Irssi__UI_processes, file, "");
        (void)newXSproto_portable("Irssi::UI::init", XS_Irssi__UI_init, file, "");
        (void)newXSproto_portable("Irssi::UI::deinit", XS_Irssi__UI_deinit, file, "");

    /* Initialisation Section */

#line 145 "UI.xs"
	irssi_boot(UI__Formats);
	irssi_boot(UI__Themes);
	irssi_boot(UI__Window);

#line 367 "UI.c"

    /* End of Initialisation Section */

#if PERL_VERSION_LE(5, 21, 5)
#  if PERL_VERSION_GE(5, 9, 0)
    if (PL_unitcheckav)
        call_list(PL_scopestack_ix, PL_unitcheckav);
#  endif
    XSRETURN_YES;
#else
    Perl_xs_boot_epilog(aTHX_ ax);
#endif
}

