/*
 * This file was generated automatically by ExtUtils::ParseXS version 3.45 from the
 * contents of Server.xs. Do not edit this file, edit Server.xs instead.
 *
 *    ANY CHANGES MADE HERE WILL BE LOST!
 *
 */

#line 1 "Server.xs"
#define PERL_NO_GET_CONTEXT
#include "module.h"
#include <irssi/src/core/misc.h>

static GSList *register_hash2list(HV *hv)
{
	HE *he;
	GSList *list;

	if (hv == NULL)
		return NULL;

	list = NULL;
	hv_iterinit(hv);
	while ((he = hv_iternext(hv)) != NULL) {
		I32 len;
		char *key = hv_iterkey(he, &len);
		int value = (int)SvIV(HeVAL(he));

		list = g_slist_append(list, g_strdup(key));
		list = g_slist_append(list, GINT_TO_POINTER(value));
	}
	return list;
}

static GSList *event_hash2list(HV *hv)
{
	HE *he;
	GSList *list;

	if (hv == NULL)
		return NULL;

	list = NULL;
	hv_iterinit(hv);
	while ((he = hv_iternext(hv)) != NULL) {
		I32 len;
		char *key = hv_iterkey(he, &len);
		char *value = SvPV_nolen(HeVAL(he));

		list = g_slist_append(list, g_strdup(key));
		list = g_slist_append(list, g_strdup(value));
	}
	return list;
}

#line 57 "Server.c"
#ifndef PERL_UNUSED_VAR
#  define PERL_UNUSED_VAR(var) if (0) var = var
#endif

#ifndef dVAR
#  define dVAR		dNOOP
#endif


/* This stuff is not part of the API! You have been warned. */
#ifndef PERL_VERSION_DECIMAL
#  define PERL_VERSION_DECIMAL(r,v,s) (r*1000000 + v*1000 + s)
#endif
#ifndef PERL_DECIMAL_VERSION
#  define PERL_DECIMAL_VERSION \
	  PERL_VERSION_DECIMAL(PERL_REVISION,PERL_VERSION,PERL_SUBVERSION)
#endif
#ifndef PERL_VERSION_GE
#  define PERL_VERSION_GE(r,v,s) \
	  (PERL_DECIMAL_VERSION >= PERL_VERSION_DECIMAL(r,v,s))
#endif
#ifndef PERL_VERSION_LE
#  define PERL_VERSION_LE(r,v,s) \
	  (PERL_DECIMAL_VERSION <= PERL_VERSION_DECIMAL(r,v,s))
#endif

/* XS_INTERNAL is the explicit static-linkage variant of the default
 * XS macro.
 *
 * XS_EXTERNAL is the same as XS_INTERNAL except it does not include
 * "STATIC", ie. it exports XSUB symbols. You probably don't want that
 * for anything but the BOOT XSUB.
 *
 * See XSUB.h in core!
 */


/* TODO: This might be compatible further back than 5.10.0. */
#if PERL_VERSION_GE(5, 10, 0) && PERL_VERSION_LE(5, 15, 1)
#  undef XS_EXTERNAL
#  undef XS_INTERNAL
#  if defined(__CYGWIN__) && defined(USE_DYNAMIC_LOADING)
#    define XS_EXTERNAL(name) __declspec(dllexport) XSPROTO(name)
#    define XS_INTERNAL(name) STATIC XSPROTO(name)
#  endif
#  if defined(__SYMBIAN32__)
#    define XS_EXTERNAL(name) EXPORT_C XSPROTO(name)
#    define XS_INTERNAL(name) EXPORT_C STATIC XSPROTO(name)
#  endif
#  ifndef XS_EXTERNAL
#    if defined(HASATTRIBUTE_UNUSED) && !defined(__cplusplus)
#      define XS_EXTERNAL(name) void name(pTHX_ CV* cv __attribute__unused__)
#      define XS_INTERNAL(name) STATIC void name(pTHX_ CV* cv __attribute__unused__)
#    else
#      ifdef __cplusplus
#        define XS_EXTERNAL(name) extern "C" XSPROTO(name)
#        define XS_INTERNAL(name) static XSPROTO(name)
#      else
#        define XS_EXTERNAL(name) XSPROTO(name)
#        define XS_INTERNAL(name) STATIC XSPROTO(name)
#      endif
#    endif
#  endif
#endif

/* perl >= 5.10.0 && perl <= 5.15.1 */


/* The XS_EXTERNAL macro is used for functions that must not be static
 * like the boot XSUB of a module. If perl didn't have an XS_EXTERNAL
 * macro defined, the best we can do is assume XS is the same.
 * Dito for XS_INTERNAL.
 */
#ifndef XS_EXTERNAL
#  define XS_EXTERNAL(name) XS(name)
#endif
#ifndef XS_INTERNAL
#  define XS_INTERNAL(name) XS(name)
#endif

/* Now, finally, after all this mess, we want an ExtUtils::ParseXS
 * internal macro that we're free to redefine for varying linkage due
 * to the EXPORT_XSUB_SYMBOLS XS keyword. This is internal, use
 * XS_EXTERNAL(name) or XS_INTERNAL(name) in your code if you need to!
 */

#undef XS_EUPXS
#if defined(PERL_EUPXS_ALWAYS_EXPORT)
#  define XS_EUPXS(name) XS_EXTERNAL(name)
#else
   /* default to internal */
#  define XS_EUPXS(name) XS_INTERNAL(name)
#endif

#ifndef PERL_ARGS_ASSERT_CROAK_XS_USAGE
#define PERL_ARGS_ASSERT_CROAK_XS_USAGE assert(cv); assert(params)

/* prototype to pass -Wmissing-prototypes */
STATIC void
S_croak_xs_usage(const CV *const cv, const char *const params);

STATIC void
S_croak_xs_usage(const CV *const cv, const char *const params)
{
    const GV *const gv = CvGV(cv);

    PERL_ARGS_ASSERT_CROAK_XS_USAGE;

    if (gv) {
        const char *const gvname = GvNAME(gv);
        const HV *const stash = GvSTASH(gv);
        const char *const hvname = stash ? HvNAME(stash) : NULL;

        if (hvname)
	    Perl_croak_nocontext("Usage: %s::%s(%s)", hvname, gvname, params);
        else
	    Perl_croak_nocontext("Usage: %s(%s)", gvname, params);
    } else {
        /* Pants. I don't think that it should be possible to get here. */
	Perl_croak_nocontext("Usage: CODE(0x%" UVxf ")(%s)", PTR2UV(cv), params);
    }
}
#undef  PERL_ARGS_ASSERT_CROAK_XS_USAGE

#define croak_xs_usage        S_croak_xs_usage

#endif

/* NOTE: the prototype of newXSproto() is different in versions of perls,
 * so we define a portable version of newXSproto()
 */
#ifdef newXS_flags
#define newXSproto_portable(name, c_impl, file, proto) newXS_flags(name, c_impl, file, proto, 0)
#else
#define newXSproto_portable(name, c_impl, file, proto) (PL_Sv=(SV*)newXS(name, c_impl, file), sv_setpv(PL_Sv, proto), (CV*)PL_Sv)
#endif /* !defined(newXS_flags) */

#if PERL_VERSION_LE(5, 21, 5)
#  define newXS_deffile(a,b) Perl_newXS(aTHX_ a,b,file)
#else
#  define newXS_deffile(a,b) Perl_newXS_deffile(aTHX_ a,b)
#endif

#line 201 "Server.c"

XS_EUPXS(XS_Irssi__Irc__Server_get_channels); /* prototype to pass -Wmissing-prototypes */
XS_EUPXS(XS_Irssi__Irc__Server_get_channels)
{
    dVAR; dXSARGS;
    if (items < 1 || items > 2)
       croak_xs_usage(cv,  "server, rejoin_channels_mode = \"\"");
    PERL_UNUSED_VAR(ax); /* -Wall */
    SP -= items;
    {
	Irssi__Irc__Server	server = irssi_ref_object(ST(0))
;
	char *	rejoin_channels_mode;
#line 55 "Server.xs"
	char *ret;
	int mode;
	SETTINGS_REC *setting;
#line 219 "Server.c"

	if (items < 2)
	    rejoin_channels_mode = "";
	else {
	    rejoin_channels_mode = (char *)SvPV_nolen(ST(1))
;
	}
#line 59 "Server.xs"
	setting = settings_get_record("rejoin_channels_on_reconnect");
	mode = strarray_find(setting->choices, rejoin_channels_mode);
	if (mode < 0)
		mode = setting->default_value.v_int;

	ret = irc_server_get_channels(server, mode);
	XPUSHs(sv_2mortal(new_pv(ret)));
	g_free(ret);
#line 236 "Server.c"
	PUTBACK;
	return;
    }
}


XS_EUPXS(XS_Irssi__Irc__Server_send_raw); /* prototype to pass -Wmissing-prototypes */
XS_EUPXS(XS_Irssi__Irc__Server_send_raw)
{
    dVAR; dXSARGS;
    if (items != 2)
       croak_xs_usage(cv,  "server, cmd");
    {
	Irssi__Irc__Server	server = irssi_ref_object(ST(0))
;
	char *	cmd = (char *)SvPV_nolen(ST(1))
;
#line 73 "Server.xs"
	irc_send_cmd(server, cmd);
#line 256 "Server.c"
    }
    XSRETURN_EMPTY;
}


XS_EUPXS(XS_Irssi__Irc__Server_send_raw_now); /* prototype to pass -Wmissing-prototypes */
XS_EUPXS(XS_Irssi__Irc__Server_send_raw_now)
{
    dVAR; dXSARGS;
    if (items != 2)
       croak_xs_usage(cv,  "server, cmd");
    {
	Irssi__Irc__Server	server = irssi_ref_object(ST(0))
;
	char *	cmd = (char *)SvPV_nolen(ST(1))
;
#line 80 "Server.xs"
	irc_send_cmd_now(server, cmd);
#line 275 "Server.c"
    }
    XSRETURN_EMPTY;
}


XS_EUPXS(XS_Irssi__Irc__Server_send_raw_first); /* prototype to pass -Wmissing-prototypes */
XS_EUPXS(XS_Irssi__Irc__Server_send_raw_first)
{
    dVAR; dXSARGS;
    if (items != 2)
       croak_xs_usage(cv,  "server, cmd");
    {
	Irssi__Irc__Server	server = irssi_ref_object(ST(0))
;
	char *	cmd = (char *)SvPV_nolen(ST(1))
;
#line 87 "Server.xs"
	irc_send_cmd_first(server, cmd);
#line 294 "Server.c"
    }
    XSRETURN_EMPTY;
}


XS_EUPXS(XS_Irssi__Irc__Server_send_raw_later); /* prototype to pass -Wmissing-prototypes */
XS_EUPXS(XS_Irssi__Irc__Server_send_raw_later)
{
    dVAR; dXSARGS;
    if (items != 2)
       croak_xs_usage(cv,  "server, cmd");
    {
	Irssi__Irc__Server	server = irssi_ref_object(ST(0))
;
	char *	cmd = (char *)SvPV_nolen(ST(1))
;
#line 94 "Server.xs"
	irc_send_cmd_later(server, cmd);
#line 313 "Server.c"
    }
    XSRETURN_EMPTY;
}


XS_EUPXS(XS_Irssi__Irc__Server_send_raw_split); /* prototype to pass -Wmissing-prototypes */
XS_EUPXS(XS_Irssi__Irc__Server_send_raw_split)
{
    dVAR; dXSARGS;
    if (items != 4)
       croak_xs_usage(cv,  "server, cmd, nickarg, max_nicks");
    {
	Irssi__Irc__Server	server = irssi_ref_object(ST(0))
;
	char *	cmd = (char *)SvPV_nolen(ST(1))
;
	int	nickarg = (int)SvIV(ST(2))
;
	int	max_nicks = (int)SvIV(ST(3))
;
#line 103 "Server.xs"
	irc_send_cmd_split(server, cmd, nickarg, max_nicks);
#line 336 "Server.c"
    }
    XSRETURN_EMPTY;
}


XS_EUPXS(XS_Irssi__Irc__Server_redirect_register); /* prototype to pass -Wmissing-prototypes */
XS_EUPXS(XS_Irssi__Irc__Server_redirect_register)
{
    dVAR; dXSARGS;
    if (items != 6)
       croak_xs_usage(cv,  "command, remote, timeout, start, stop, opt");
    {
	char *	command = (char *)SvPV_nolen(ST(0))
;
	int	remote = (int)SvIV(ST(1))
;
	int	timeout = (int)SvIV(ST(2))
;
	SV *	start = ST(3)
;
	SV *	stop = ST(4)
;
	SV *	opt = ST(5)
;
#line 116 "Server.xs"
	server_redirect_register_list(command, remote, timeout, register_hash2list(hvref(start)),
	                              register_hash2list(hvref(stop)),
	                              register_hash2list(hvref(opt)), 1);
#line 365 "Server.c"
    }
    XSRETURN_EMPTY;
}


XS_EUPXS(XS_Irssi__Irc__Server_redirect_event); /* prototype to pass -Wmissing-prototypes */
XS_EUPXS(XS_Irssi__Irc__Server_redirect_event)
{
    dVAR; dXSARGS;
    if (items != 7)
       croak_xs_usage(cv,  "server, command, count, arg, remote, failure_signal, signals");
    {
	Irssi__Irc__Server	server = irssi_ref_object(ST(0))
;
	char *	command = (char *)SvPV_nolen(ST(1))
;
	int	count = (int)SvIV(ST(2))
;
	char *	arg = (char *)SvPV_nolen(ST(3))
;
	int	remote = (int)SvIV(ST(4))
;
	char *	failure_signal = (char *)SvPV_nolen(ST(5))
;
	SV *	signals = ST(6)
;
#line 130 "Server.xs"
	server_redirect_event_list(server, command, count, *arg == '\0' ? NULL : arg, remote,
				   *failure_signal == '\0' ? NULL : failure_signal,
				   event_hash2list(hvref(signals)));
#line 396 "Server.c"
    }
    XSRETURN_EMPTY;
}


XS_EUPXS(XS_Irssi__Irc__Server_redirect_get_signal); /* prototype to pass -Wmissing-prototypes */
XS_EUPXS(XS_Irssi__Irc__Server_redirect_get_signal)
{
    dVAR; dXSARGS;
    if (items != 4)
       croak_xs_usage(cv,  "server, prefix, event, args");
    {
	Irssi__Irc__Server	server = irssi_ref_object(ST(0))
;
	char *	prefix = (char *)SvPV_nolen(ST(1))
;
	char *	event = (char *)SvPV_nolen(ST(2))
;
	char *	args = (char *)SvPV_nolen(ST(3))
;
	char *	RETVAL;
	dXSTARG;
#line 141 "Server.xs"
	RETVAL = (char *) server_redirect_get_signal(server, prefix, event, args);
#line 421 "Server.c"
	sv_setpv(TARG, RETVAL);
	XSprePUSH;
	PUSHTARG;
    }
    XSRETURN(1);
}


XS_EUPXS(XS_Irssi__Irc__Server_redirect_peek_signal); /* prototype to pass -Wmissing-prototypes */
XS_EUPXS(XS_Irssi__Irc__Server_redirect_peek_signal)
{
    dVAR; dXSARGS;
    if (items != 4)
       croak_xs_usage(cv,  "server, prefix, event, args");
    {
	Irssi__Irc__Server	server = irssi_ref_object(ST(0))
;
	char *	prefix = (char *)SvPV_nolen(ST(1))
;
	char *	event = (char *)SvPV_nolen(ST(2))
;
	char *	args = (char *)SvPV_nolen(ST(3))
;
#line 152 "Server.xs"
	int redirection;
#line 447 "Server.c"
	char *	RETVAL;
	dXSTARG;
#line 154 "Server.xs"
	RETVAL = (char *) server_redirect_peek_signal(server, prefix, event, args, &redirection);
#line 452 "Server.c"
	sv_setpv(TARG, RETVAL);
	XSprePUSH;
	PUSHTARG;
    }
    XSRETURN(1);
}


XS_EUPXS(XS_Irssi__Irc__Server_isupport); /* prototype to pass -Wmissing-prototypes */
XS_EUPXS(XS_Irssi__Irc__Server_isupport)
{
    dVAR; dXSARGS;
    if (items != 2)
       croak_xs_usage(cv,  "server, name");
    {
	Irssi__Irc__Server	server = irssi_ref_object(ST(0))
;
	char *	name = (char *)SvPV_nolen(ST(1))
;
	char *	RETVAL;
	dXSTARG;
#line 163 "Server.xs"
	RETVAL = (char *) g_hash_table_lookup(server->isupport, name);
#line 476 "Server.c"
	sv_setpv(TARG, RETVAL);
	XSprePUSH;
	PUSHTARG;
    }
    XSRETURN(1);
}


XS_EUPXS(XS_Irssi__Irc__Server_irc_server_cap_toggle); /* prototype to pass -Wmissing-prototypes */
XS_EUPXS(XS_Irssi__Irc__Server_irc_server_cap_toggle)
{
    dVAR; dXSARGS;
    if (items != 3)
       croak_xs_usage(cv,  "server, cap, enable");
    {
	Irssi__Irc__Server	server = irssi_ref_object(ST(0))
;
	char *	cap = (char *)SvPV_nolen(ST(1))
;
	int	enable = (int)SvIV(ST(2))
;
	int	RETVAL;
	dXSTARG;
#line 173 "Server.xs"
	RETVAL = irc_cap_toggle(server, cap, enable);
#line 502 "Server.c"
	XSprePUSH;
	PUSHi((IV)RETVAL);
    }
    XSRETURN(1);
}

#ifdef __cplusplus
extern "C"
#endif
XS_EXTERNAL(boot_Irssi__Irc__Server); /* prototype to pass -Wmissing-prototypes */
XS_EXTERNAL(boot_Irssi__Irc__Server)
{
#if PERL_VERSION_LE(5, 21, 5)
    dVAR; dXSARGS;
#else
    dVAR; dXSBOOTARGSXSAPIVERCHK;
#endif
#if PERL_VERSION_LE(5, 8, 999) /* PERL_VERSION_LT is 5.33+ */
    char* file = __FILE__;
#else
    const char* file = __FILE__;
#endif

    PERL_UNUSED_VAR(file);

    PERL_UNUSED_VAR(cv); /* -W */
    PERL_UNUSED_VAR(items); /* -W */
#if PERL_VERSION_LE(5, 21, 5)
    XS_VERSION_BOOTCHECK;
#  ifdef XS_APIVERSION_BOOTCHECK
    XS_APIVERSION_BOOTCHECK;
#  endif
#endif

        (void)newXSproto_portable("Irssi::Irc::Server::get_channels", XS_Irssi__Irc__Server_get_channels, file, "$;$");
        (void)newXSproto_portable("Irssi::Irc::Server::send_raw", XS_Irssi__Irc__Server_send_raw, file, "$$");
        (void)newXSproto_portable("Irssi::Irc::Server::send_raw_now", XS_Irssi__Irc__Server_send_raw_now, file, "$$");
        (void)newXSproto_portable("Irssi::Irc::Server::send_raw_first", XS_Irssi__Irc__Server_send_raw_first, file, "$$");
        (void)newXSproto_portable("Irssi::Irc::Server::send_raw_later", XS_Irssi__Irc__Server_send_raw_later, file, "$$");
        (void)newXSproto_portable("Irssi::Irc::Server::send_raw_split", XS_Irssi__Irc__Server_send_raw_split, file, "$$$$");
        (void)newXSproto_portable("Irssi::Irc::Server::redirect_register", XS_Irssi__Irc__Server_redirect_register, file, "$$$$$$");
        (void)newXSproto_portable("Irssi::Irc::Server::redirect_event", XS_Irssi__Irc__Server_redirect_event, file, "$$$$$$$");
        (void)newXSproto_portable("Irssi::Irc::Server::redirect_get_signal", XS_Irssi__Irc__Server_redirect_get_signal, file, "$$$$");
        (void)newXSproto_portable("Irssi::Irc::Server::redirect_peek_signal", XS_Irssi__Irc__Server_redirect_peek_signal, file, "$$$$");
        (void)newXSproto_portable("Irssi::Irc::Server::isupport", XS_Irssi__Irc__Server_isupport, file, "$$");
        (void)newXSproto_portable("Irssi::Irc::Server::irc_server_cap_toggle", XS_Irssi__Irc__Server_irc_server_cap_toggle, file, "$$$");
#if PERL_VERSION_LE(5, 21, 5)
#  if PERL_VERSION_GE(5, 9, 0)
    if (PL_unitcheckav)
        call_list(PL_scopestack_ix, PL_unitcheckav);
#  endif
    XSRETURN_YES;
#else
    Perl_xs_boot_epilog(aTHX_ ax);
#endif
}

